@@ -1,4 +1,3 @@
-/* $Xorg: xprop.c,v 1.6 2001/02/09 02:05:56 xorgcvs Exp $ */
 /*
 
 Copyright 1990, 1998  The Open Group
@@ -27,7 +26,6 @@ other dealings in this Software without prior written authorization
 from The Open Group.
 
 */
-/* $XFree86: xc/programs/xprop/xprop.c,v 1.15 2003/09/24 02:43:38 dawes Exp $ */
 
 #include "config.h"
 
@@ -35,6 +33,7 @@ from The Open Group.
 #include <X11/Xos.h>
 #include <X11/Xfuncs.h>
 #include <X11/Xutil.h>
+#include <sys/ioctl.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <ctype.h>
@@ -67,6 +66,8 @@ from The Open Group.
 /* isprint() in "C" locale */
 #define c_isprint(c) ((c) >= 0x20 && (c) < 0x7f)
 
+static int term_width = 144 + 8;
+
 /*
  *
  * The Thunk Manager - routines to create, add to, and free thunk lists
@@ -88,21 +89,15 @@ Create_Thunk_List (void)
 {
     thunk *tptr;
 
-    tptr = (thunk *) Malloc(sizeof(thunk));
+    tptr = malloc(sizeof(thunk));
+    if (!tptr)
+	Fatal_Error("Out of memory!");
 
     tptr->thunk_count = 0;
 
     return tptr;
 }
 
-#ifdef notused
-static void
-Free_Thunk_List (thunk *list)
-{
-    free(list);
-}
-#endif
-
 static thunk *
 Add_Thunk (thunk *list, thunk t)
 {
@@ -110,7 +105,7 @@ Add_Thunk (thunk *list, thunk t)
 
     i = list->thunk_count;
 
-    list = (thunk *) realloc(list, (i+1)*sizeof(thunk));
+    list = realloc(list, (i+1)*sizeof(thunk));
     if (!list)
 	Fatal_Error("Out of memory!");
 
@@ -124,20 +119,6 @@ Add_Thunk (thunk *list, thunk t)
  * Misc. routines
  */
 
-static char *
-Copy_String (const char *string)
-{
-    char *new;
-    int length;
-
-    length = strlen(string) + 1;
-
-    new = (char *) Malloc(length);
-    memcpy(new, string, length);
-
-    return new;
-}
-
 static int
 Read_Char (FILE *stream)
 {
@@ -182,13 +163,17 @@ Read_Quoted (FILE *stream)
 	    c = Read_Char(stream);
 	    if (c == '\n') {
 		ptr--; length++;
-	    } else
+	    } else {
 		ptr++[0] = c; length--;
+	    }
 	}
     }
     ptr++[0] = '\0';
 
-    return Copy_String(_large_buffer);
+    ptr = strdup(_large_buffer);
+    if (!ptr)
+	Fatal_Error("Out of memory!");
+    return ptr;
 }
 
 /*
@@ -298,7 +283,7 @@ Lookup_Formats (Atom atom, const char **format, const char **dformat)
 static void
 Add_Mapping (Atom atom, const char *format, const char *dformat)
 {
-    thunk t;
+    thunk t = {0};
 
     if (!_property_formats)
 	_property_formats = Create_Thunk_List();
@@ -409,6 +394,7 @@ static propertyRec windowPropTable[] = {
     {"RECTANGLE",	XA_RECTANGLE,	 "16iicc",    RECTANGLE_DFORMAT },
     {"RGB_COLOR_MAP",	XA_RGB_COLOR_MAP,"32xcccccccxx",RGB_COLOR_MAP_DFORMAT},
     {"STRING",		XA_STRING,	 "8s",	      0 },
+    {"UTF8_STRING",		0,	 "8u",	      0 },
     {"WINDOW",		XA_WINDOW,	 "32x",	      ": window id # $0+\n" },
     {"VISUALID",	XA_VISUALID,	 "32x",	      ": visual id # $0\n" },
     {"WM_COLORMAP_WINDOWS",	0,	 "32x",       ": window id # $0+\n"},
@@ -547,7 +533,8 @@ static void
 Read_Mappings (FILE *stream)
 {
     char format_buffer[100];
-    char name[1000], *dformat, *format;
+    char name[1000];
+    const char *dformat, *format;
     int count, c;
     Atom atom;
 
@@ -556,7 +543,9 @@ Read_Mappings (FILE *stream)
 	    Fatal_Error("Bad format file format.");
 
 	atom = Parse_Atom(name, False);
-	format = Copy_String(format_buffer);
+	format = strdup(format_buffer);
+	if (!format)
+	    Fatal_Error("Out of memory!");
 
 	Read_White_Space(stream);
 	dformat = D_DFORMAT;
@@ -683,7 +672,7 @@ _put_char (char c)
 }
 
 static void
-_format_char (char c)
+_format_char (char c, int unicode)
 {
     switch (c) {
       case '\\':
@@ -701,17 +690,21 @@ _format_char (char c)
 	break;
       default:
 	if (!c_isprint(c)) {
-	    _put_char('\\');
-	    snprintf(_buf_ptr, _buf_len, "%03o", (unsigned char) c);
-	    _buf_ptr += 3;
-	    _buf_len -= 3;
+	    if (unicode && (c & 0x80)) {
+		_put_char(c);
+	    } else {
+		_put_char('\\');
+		snprintf(_buf_ptr, _buf_len, "%03o", (unsigned char) c);
+		_buf_ptr += 3;
+		_buf_len -= 3;
+	    }
 	} else
 	  _put_char(c);
     }
 }
 
 static const char *
-Format_String (const char *string)
+Format_String (const char *string, int unicode)
 {
     char c;
 
@@ -720,7 +713,7 @@ Format_String (const char *string)
     _put_char('\"');
 
     while ((c = string++[0]))
-	_format_char(c);
+	_format_char(c, unicode);
 
     *_buf_ptr++ = '"';
     *_buf_ptr++ = '\0';
@@ -728,17 +721,19 @@ Format_String (const char *string)
 }
 
 static const char *
-Format_Len_String (const char *string, int len)
+Format_Len_String (const char *string, int len, int unicode)
 {
     char *data;
     const char *result;
 
-    data = (char *) Malloc(len+1);
+    data = malloc(len+1);
+    if (!data)
+	Fatal_Error("Out of memory!");
 
     memcpy(data, string, len);
     data[len] = '\0';
 
-    result = Format_String(data);
+    result = Format_String(data, unicode);
     free(data);
 
     return result;
@@ -756,30 +751,59 @@ is_utf8_locale (void)
 #endif
 }
 
+static int
+is_truecolor_term (void)
+{
+    char *colorterm = getenv( "COLORTERM" );
+
+    if (colorterm && !strcmp(colorterm,"truecolor"))
+	return 1;
+
+    return 0;
+}
+
 static const char *
 Format_Icons (const unsigned long *icon, int len)
 {
-    char *result = NULL, *tail = NULL;
+    static char *result = NULL;
+    char *tail = NULL;
     int alloced;
     const unsigned long *end = icon + len / sizeof (unsigned long);
 
+    free(result);
+    result = NULL;
+
     alloced = 0;
 
     while (icon < end)
     {
-	unsigned long width, height;
+	unsigned long width, height, display_width;
+	unsigned int icon_pixel_bytes;
+	unsigned int icon_line_bytes;
 	int w, h;
 	int offset;
 	
 	width = *icon++;
 	height = *icon++;
+	display_width = width * 2; /* Two characters per icon pixel. */
+
+	icon_pixel_bytes = 1;
+	if (is_truecolor_term())
+	    icon_pixel_bytes = 25; /* 16 control characters, and up to 9 chars of RGB. */
+	else if (is_utf8_locale())
+	    icon_pixel_bytes = 3; /* Up to 3 bytes per character in that mode. */
+
+	/* Initial tab, pixels, and newline. */
+	icon_line_bytes = 8 + display_width * icon_pixel_bytes + 1;
 
 	offset = (tail - result);
 	
-	alloced += 80;				/* For the header */
-	alloced += (width*4 + 8) * height;	/* For the rows (plus padding) */
+	alloced += 80;				/* For the header, final newline, color reset */
+	alloced += icon_line_bytes * height;	/* For the rows */
 	
-	result = Realloc (result, alloced);
+	result = realloc (result, alloced);
+	if (!result)
+	    Fatal_Error("Out of memory!");
 	tail = &result[offset];
 
 	if (end - icon < width * height)
@@ -787,9 +811,9 @@ Format_Icons (const unsigned long *icon, int len)
 
 	tail += sprintf (tail, "\tIcon (%lu x %lu):\n", width, height);
 
-	if (width > 144 || height > 144)
+	if ((display_width + 8) > term_width || height > 144)
 	{
-	    tail += sprintf (tail, "\t(not shown)");
+	    tail += sprintf (tail, "\t(not shown)\n");
 	    icon += width * height;
 	    continue;
 	}
@@ -814,7 +838,17 @@ Format_Icons (const unsigned long *icon, int len)
 					   (587 * (g / 255.0)) +
 					   (114 * (b / 255.0))));
 
-		if (is_utf8_locale())
+		if (is_truecolor_term())
+		{
+		    float opacity = a / 255.0;
+
+		    r = r * opacity;
+		    g = g * opacity;
+		    b = b * opacity;
+
+		    tail += sprintf (tail, "\033[38;2;%d;%d;%dm\342\226\210\342\226\210", r, g, b );
+		}
+		else if (is_utf8_locale())
 		{
 		    static const char palette[][4] =
 		    {
@@ -828,7 +862,7 @@ Format_Icons (const unsigned long *icon, int len)
 
 		    idx = (brightness * ((sizeof (palette)/sizeof(palette[0])) - 1)) / 1000;
 
-		    tail += sprintf (tail, "%s", palette[idx]);
+		    tail += sprintf (tail, "%s%s", palette[idx], palette[idx]);
 		}
 		else
 		{
@@ -839,12 +873,17 @@ Format_Icons (const unsigned long *icon, int len)
 		    idx = (brightness * (sizeof(palette) - 2)) / 1000;
 		    
 		    *tail++ = palette[idx];
+		    *tail++ = palette[idx];
 		}
 	    }
 
 	    tail += sprintf (tail, "\n");
 	}
 
+	/* Reset colors. */
+	if (is_truecolor_term())
+	    tail += sprintf (tail, "\033[0m");
+
 	tail += sprintf (tail, "\n");
     }
 
@@ -855,7 +894,7 @@ static const char *
 Format_Len_Text (const char *string, int len, Atom encoding)
 {
     XTextProperty textprop;
-    char **list;
+    char **list, **start_list;
     int count;
 
     /* Try to convert to local encoding. */
@@ -863,7 +902,8 @@ Format_Len_Text (const char *string, int len, Atom encoding)
     textprop.format = 8;
     textprop.value = (unsigned char *) string;
     textprop.nitems = len;
-    if (XmbTextPropertyToTextList(dpy, &textprop, &list, &count) == Success) {
+    if (XmbTextPropertyToTextList(dpy, &textprop, &start_list, &count) == Success) {
+	list = start_list;
 	_buf_ptr = _formatting_buffer;
 	_buf_len = MAXSTR;
 	*_buf_ptr++ = '"';
@@ -897,11 +937,127 @@ Format_Len_Text (const char *string, int len, Atom encoding)
 		_buf_len -= 4;
 	    }
 	}
+	XFreeStringList(start_list);
 	*_buf_ptr++ = '"';
 	*_buf_ptr++ = '\0';
 	return _formatting_buffer;
     } else
-	return Format_Len_String(string, len);
+	return Format_Len_String(string, len, 0);
+}
+
+/*
+ * Validate a string as UTF-8 encoded according to RFC 3629
+ *
+ * Simply, a unicode code point (up to 21-bits long) is encoded as follows:
+ *
+ *    Char. number range  |        UTF-8 octet sequence
+ *       (hexadecimal)    |              (binary)
+ *    --------------------+---------------------------------------------
+ *    0000 0000-0000 007F | 0xxxxxxx
+ *    0000 0080-0000 07FF | 110xxxxx 10xxxxxx
+ *    0000 0800-0000 FFFF | 1110xxxx 10xxxxxx 10xxxxxx
+ *    0001 0000-0010 FFFF | 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
+ *
+ * Validation is done left-to-right, and an error condition, if any, refers to
+ * only the left-most problem in the string.
+ *
+ * Return values:
+ *   UTF8_VALID: Valid UTF-8 encoded string
+ *   UTF8_OVERLONG: Using more bytes than needed for a code point
+ *   UTF8_SHORT_TAIL: Not enough bytes in a multi-byte sequence
+ *   UTF8_LONG_TAIL: Too many bytes in a multi-byte sequence
+ *   UTF8_FORBIDDEN_VALUE: Forbidden prefix or code point outside 0x10FFFF
+ */
+#define UTF8_VALID 0
+#define UTF8_FORBIDDEN_VALUE 1
+#define UTF8_OVERLONG 2
+#define UTF8_SHORT_TAIL 3
+#define UTF8_LONG_TAIL 4
+static int
+is_valid_utf8 (const char *string, int len)
+{
+    unsigned long codepoint = 0;
+    int rem, i;
+    unsigned char c;
+
+    rem = 0;
+    for (i = 0; i < len; i++) {
+	c = (unsigned char) string[i];
+
+	/* Order of type check:
+	 *   - Single byte code point
+	 *   - Non-starting byte of multi-byte sequence
+	 *   - Start of 2-byte sequence
+	 *   - Start of 3-byte sequence
+	 *   - Start of 4-byte sequence
+	 */
+	if (!(c & 0x80)) {
+	    if (rem > 0) return UTF8_SHORT_TAIL;
+	    rem = 0;
+	    codepoint = c;
+	} else if ((c & 0xC0) == 0x80) {
+	    if (rem == 0) return UTF8_LONG_TAIL;
+	    rem--;
+	    codepoint |= (c & 0x3F) << (rem * 6);
+	    if (codepoint == 0) return UTF8_OVERLONG;
+	} else if ((c & 0xE0) == 0xC0) {
+	    if (rem > 0) return UTF8_SHORT_TAIL;
+	    rem = 1;
+	    codepoint = (c & 0x1F) << 6;
+	    if (codepoint == 0) return UTF8_OVERLONG;
+	} else if ((c & 0xF0) == 0xE0) {
+	    if (rem > 0) return UTF8_SHORT_TAIL;
+	    rem = 2;
+	    codepoint = (c & 0x0F) << 12;
+	} else if ((c & 0xF8) == 0xF0) {
+	    if (rem > 0) return UTF8_SHORT_TAIL;
+	    rem = 3;
+	    codepoint = (c & 0x07) << 18;
+	    if (codepoint > 0x10FFFF) return UTF8_FORBIDDEN_VALUE;
+	} else
+	    return UTF8_FORBIDDEN_VALUE;
+    }
+
+    return UTF8_VALID;
+}
+
+static const char *
+Format_Len_Unicode (const char *string, int len)
+{
+    char *data;
+    const char *result, *error;
+
+    int validity = is_valid_utf8(string, len);
+
+    if (validity != UTF8_VALID) {
+	switch (validity) {
+	  case UTF8_FORBIDDEN_VALUE:
+	    error = "<Invalid UTF-8 string: Forbidden value> "; break;
+	  case UTF8_OVERLONG:
+	    error = "<Invalid UTF-8 string: Overlong encoding> "; break;
+	  case UTF8_SHORT_TAIL:
+	    error = "<Invalid UTF-8 string: Tail too short> "; break;
+	  case UTF8_LONG_TAIL:
+	    error = "<Invalid UTF-8 string: Tail too long> "; break;
+	  default:
+	    error = "<Invalid UTF-8 string: Unknown error>"; break;
+	}
+
+	result = Format_Len_String(string, len, 0);
+	/* result is stored in _formatting_buffer, so make a temporary
+	   copy before we overwrite _formatting_buffer with error */
+	data = strdup(result);
+	if (!data)
+	    Fatal_Error("Out of memory!");
+
+	memcpy(_formatting_buffer, error, strlen(error)+1);
+	strcat(_formatting_buffer, data);
+	free(data);
+
+	return _formatting_buffer;
+    }
+
+    return Format_Len_String(string, len, is_utf8_locale());
 }
 
 /*
@@ -955,7 +1111,9 @@ Format_Thunk (thunk t, char format_char)
 
     switch (format_char) {
       case 's':
-	return Format_Len_String(t.extra_value, (int)t.value);
+	return Format_Len_String(t.extra_value, (int)t.value, 0);
+      case 'u':
+	return Format_Len_Unicode(t.extra_value, (int)t.value);
       case 't':
 	return Format_Len_Text(t.extra_value, (int)t.value, t.extra_encoding);
       case 'x':
@@ -1243,7 +1401,7 @@ static thunk *
 Break_Down_Property (const char *pointer, int length, Atom type, const char *format, int size)
 {
     thunk *thunks;
-    thunk t;
+    thunk t = {0};
     int i;
     char format_char;
 
@@ -1252,7 +1410,7 @@ Break_Down_Property (const char *pointer, int length, Atom type, const char *for
 
     while (length >= size/8) {
 	format_char = Get_Format_Char(format, i);
-	if (format_char == 's')
+	if (format_char == 's' || format_char == 'u')
 	    t.value = Extract_Len_String(&pointer,&length,size,&t.extra_value);
 	else if (format_char == 't') {
 	    t.extra_encoding = type;
@@ -1313,9 +1471,15 @@ Get_Window_Property_Data_And_Type (Atom atom,
     unsigned long nitems;
     unsigned long nbytes;
     unsigned long bytes_after;
-    unsigned char *prop;
+    static unsigned char *prop = NULL;
     int status;
-	
+
+    if (prop)
+    {
+	XFree(prop);
+	prop = NULL;
+    }
+
     status = XGetWindowProperty(dpy, target_win, atom, 0, (max_len+3)/4,
 				False, AnyPropertyType, &actual_type,
 				&actual_format, &nitems, &bytes_after,
@@ -1388,8 +1552,8 @@ Show_Prop (const char *format, const char *dformat, const char *prop)
     }
 
     thunks = Break_Down_Property(data, (int)length, type, format, size);
-
     Display_Property(thunks, dformat, format);
+    free(thunks);
 }
 
 static void
@@ -1405,6 +1569,7 @@ Show_All_Props (void)
 	    name = Format_Atom(atoms[i]);
 	    Show_Prop(NULL, NULL, name);
 	}
+	XFree(atoms);
     } else
 	for (i = 0; i < font->n_properties; i++) {
 	    atom = font->properties[i].name;
@@ -1417,7 +1582,7 @@ static thunk *
 Handle_Prop_Requests (int argc, char **argv)
 {
     char *format, *dformat, *prop;
-    thunk *thunks, t;
+    thunk *thunks, t = {0};
 
     /* if no prop referenced, by default list all properties for given window */
     if (!argc) {
@@ -1434,11 +1599,11 @@ Handle_Prop_Requests (int argc, char **argv)
 	/* Get overriding formats, if any */
 	if (Is_A_Format(argv[0])) {
 	    format = argv++[0]; argc--;
-	    if (!argc) usage();
+	    if (!argc) usage("format specified without atom");
 	}
 	if (Is_A_DFormat(argv[0])) {
 	    dformat = argv++[0]; argc--;
-	    if (!argc) usage();
+	    if (!argc) usage("dformat specified without atom");
 	}
 
 	/* Get property name */
@@ -1477,7 +1642,7 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
     int size;
     char format_char;
     Atom type = 0;
-    unsigned char *data = NULL;
+    const unsigned char *data = NULL;
     int nelements = 0;
 
     atom = Parse_Atom(propname, False);
@@ -1495,7 +1660,14 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	if (size != 8)
 	    Fatal_Error("can't use format character 's' with any size except 8.");
 	type = XA_STRING;
-	data = (unsigned char *) value;
+	data = (const unsigned char *) value;
+	nelements = strlen(value);
+	break;
+      case 'u':
+	if (size != 8)
+	    Fatal_Error("can't use format character 'u' with any size except 8.");
+	type = XInternAtom(dpy, "UTF8_STRING", False);
+	data = (const unsigned char *) value;
 	nelements = strlen(value);
 	break;
       case 't': {
@@ -1524,11 +1696,11 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	intvalue = strtoul(tmp, NULL, 0);
 	switch(size) {
 	    case 8:
-	        data8[0] = intvalue; data = (unsigned char *) data8; break;
+	        data8[0] = intvalue; data = (const unsigned char *) data8; break;
 	    case 16:    
-	        data16[0] = intvalue; data = (unsigned char *) data16; break;
+	        data16[0] = intvalue; data = (const unsigned char *) data16; break;
 	    case 32:    
-	        data32[0] = intvalue; data = (unsigned char *) data32; break;
+	        data32[0] = intvalue; data = (const unsigned char *) data32; break;
 	}
 	tmp = strtok(NULL,",");
 	while(tmp != NULL){
@@ -1564,11 +1736,11 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	intvalue = strtoul(tmp, NULL, 0);
 	switch(size) {
 	    case 8:
-	        data8[0] = intvalue; data = (unsigned char *) data8; break;
+	        data8[0] = intvalue; data = (const unsigned char *) data8; break;
 	    case 16:    
-	        data16[0] = intvalue; data = (unsigned char *) data16; break;
+	        data16[0] = intvalue; data = (const unsigned char *) data16; break;
 	    case 32:    
-	        data32[0] = intvalue; data = (unsigned char *) data32; break;
+	        data32[0] = intvalue; data = (const unsigned char *) data32; break;
 	}
 	tmp = strtok(NULL,",");
 	while(tmp != NULL){
@@ -1609,11 +1781,11 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	type = XA_INTEGER;
 	switch (size) {
 	  case 8:
-	    data8 = boolvalue; data = (unsigned char *) &data8; break;
+	    data8 = boolvalue; data = (const unsigned char *) &data8; break;
 	  case 16:
-	    data16 = boolvalue; data = (unsigned char *) &data16; break;
+	    data16 = boolvalue; data = (const unsigned char *) &data16; break;
 	  case 32: default:
-	    data32 = boolvalue; data = (unsigned char *) &data32; break;
+	    data32 = boolvalue; data = (const unsigned char *) &data32; break;
 	}
 	nelements = 1;
 	break;
@@ -1622,7 +1794,7 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	static Atom avalue;
 	avalue = Parse_Atom(value, False);
 	type = XA_ATOM;
-	data = (unsigned char *) &avalue;
+	data = (const unsigned char *) &avalue;
 	nelements = 1;
 	break;
       }
@@ -1643,10 +1815,11 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
  */
 
 void
-usage (void)
+print_help (void)
 {
-    static const char help_message[] = 
+    static const char *help_message =
 "where options include:\n"
+"    -help                          print out a summary of command line options\n"
 "    -grammar                       print out full grammar for command line\n"
 "    -display host:dpy              the X server to contact\n"
 "    -id id                         resource id of window to examine\n"
@@ -1660,14 +1833,30 @@ usage (void)
 "    -fs filename                   where to look for formats for properties\n"
 "    -frame                         don't ignore window manager frames\n"
 "    -f propname format [dformat]   formats to use for property of given name\n"
-"    -spy                           examine window properties forever\n";
+"    -spy                           examine window properties forever\n"
+"    -version                       print program version\n";
 
 
     fflush (stdout);
+
     fprintf (stderr,
 	     "usage:  %s [-options ...] [[format [dformat]] atom] ...\n\n", 
 	     program_name);
     fprintf (stderr, "%s\n", help_message);
+}
+
+void help (void) {
+	print_help();
+	exit(0);
+}
+
+void
+usage (const char *errmsg)
+{
+    if (errmsg != NULL)
+	fprintf (stderr, "%s: %s\n\n", program_name, errmsg);
+
+    print_help();
     exit (1);
 }
 
@@ -1701,7 +1890,7 @@ Parse_Format_Mapping (int *argc, char ***argv)
 #define ARGC (*argc)
 #define ARGV (*argv)
 #define OPTION ARGV[0]
-#define NXTOPT if (++ARGV, --ARGC==0) usage()
+#define NXTOPT if (++ARGV, --ARGC==0) usage("insufficent arguments for -format")
     char *type_name, *format, *dformat;
   
     NXTOPT; type_name = OPTION;
@@ -1754,6 +1943,13 @@ main (int argc, char **argv)
     int n;
     char **nargv;
 
+#ifdef TIOCGWINSZ
+    struct winsize ws;
+    ws.ws_col = 0;
+    if (ioctl(STDIN_FILENO, TIOCGWINSZ, &ws) != -1 && ws.ws_col != 0)
+	term_width = ws.ws_col;
+#endif
+
     INIT_NAME;
 
     /* Set locale for XmbTextProptertyToTextList and iswprint(). */
@@ -1784,6 +1980,10 @@ main (int argc, char **argv)
     while (argv++, --argc>0 && **argv == '-') {
 	if (!strcmp(argv[0], "-"))
 	    continue;
+	if (!strcmp(argv[0], "-help")) {
+	    help ();
+	    /* NOTREACHED */
+	}
 	if (!strcmp(argv[0], "-grammar")) {
 	    grammar ();
 	    /* NOTREACHED */
@@ -1797,12 +1997,12 @@ main (int argc, char **argv)
 	    continue;
 	}
 	if (!strcmp(argv[0], "-len")) {
-	    if (++argv, --argc == 0) usage();
+	    if (++argv, --argc == 0) usage("-len requires an argument");
 	    max_len = atoi(argv[0]);
 	    continue;
 	}
 	if (!strcmp(argv[0], "-formats") || !strcmp(argv[0], "-fs")) {
-	    if (++argv, --argc == 0) usage();
+	    if (++argv, --argc == 0) usage("-fs requires an argument");
 	    if (!(stream = fopen(argv[0], "r")))
 		Fatal_Error("unable to open file %s for reading.", argv[0]);
 	    Read_Mappings(stream);
@@ -1810,22 +2010,22 @@ main (int argc, char **argv)
 	    continue;
 	}
 	if (!strcmp(argv[0], "-font")) {
-	    if (++argv, --argc == 0) usage();
+	    if (++argv, --argc == 0) usage("-font requires an argument");
 	    font = Open_Font(argv[0]);
 	    target_win = -1;
 	    continue;
 	}
 	if (!strcmp(argv[0], "-remove")) {
-	    thunk t;
-	    if (++argv, --argc == 0) usage();
+	    thunk t = {0};
+	    if (++argv, --argc == 0) usage("-remove requires an argument");
 	    t.propname = argv[0];
 	    if (remove_props == NULL) remove_props = Create_Thunk_List();
 	    remove_props = Add_Thunk(remove_props, t);
 	    continue;
 	}
 	if (!strcmp(argv[0], "-set")) {
-	    thunk t;
-	    if (argc < 3) usage();
+	    thunk t = {0};
+	    if (argc < 3) usage("insufficent arguments for -set");
 	    t.propname = argv[1];
 	    t.extra_value = argv[2];
 	    argv += 3; argc -= 3;
@@ -1841,11 +2041,20 @@ main (int argc, char **argv)
 	    Parse_Format_Mapping(&argc, &argv);
 	    continue;
 	}
-	usage();
+	if (!strcmp(argv[0], "-version")) {
+	    puts(PACKAGE_STRING);
+	    exit(0);
+	}
+	fprintf (stderr, "%s: unrecognized argument %s\n\n",
+		 program_name, argv[0]);
+	usage(NULL);
     }
 
-    if ((remove_props != NULL || set_props != NULL) && argc > 0)
-	usage();
+    if ((remove_props != NULL || set_props != NULL) && argc > 0) {
+	fprintf (stderr, "%s: unrecognized argument %s\n\n",
+		 program_name, argv[0]);
+	usage(NULL);
+    }
 
     if (target_win == None)
 	target_win = Select_Window(dpy, !frame_only);
